
from __future__ import print_function

import _setup_test_env  # noqa
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_greater
import time
import sys
from pprint import pprint

from returnn.theano.engine_task import TaskThread, TrainTaskThread, EvalTaskThread
from returnn.theano.device import Device
from returnn.config import Config
from returnn.log import log
from returnn.util.basic import hms, NumbersDict


config = Config()
config.update({
  "multiprocessing": False,
  "blocking": True,
  "device": "cpu",
  "num_epochs": 1,
  "num_inputs": 3,
  "num_outputs": 2,
})
config.network_topology_json = """
{
"output": {"class": "softmax", "loss": "ce"}
}
"""

class DummyBatches:
  def __init__(self):
    pass

  def completed_frac(self):
    return 0.5

  def get_current_batch_idx(self):
    return 0


class DummyNetwork:
  recurrent = True
  ctc_priors = None
  loss = "ce"


class DummyUpdater:
  isInitialized = True


def getDeviceBatchRunParent(dev, task):

  if task == "train":
    parent_clazz_base = TrainTaskThread
  elif task == "eval":
    parent_clazz_base = EvalTaskThread
  else:
    assert "invalid: %r" % task
    return

  class DummyDeviceBatchRunParent(parent_clazz_base):

    def __init__(self):
      init_kwargs = {}
      if parent_clazz_base is TrainTaskThread:
        init_kwargs.update(dict(
          learning_rate=1.0,
          updater=DummyUpdater()
        ))
      super(DummyDeviceBatchRunParent, self).__init__(
        network=DummyNetwork(),
        report_prefix="test %s run" % task,
        batches=DummyBatches(),
        devices=[dev],
        data=None,
        **init_kwargs
      )

      self.interactive = False
      self.start_time = time.time()

    def start(self):
      # Do nothing. Don't start the thread.
      print("DummyDeviceBatchRunParent: dummy start thread (ignored)")

  return DummyDeviceBatchRunParent()


class DummyDevice(Device):
  def __init__(self):
    super(DummyDevice, self).__init__("cpu", config=config, blocking=True)

  def run(self, task):
    print("DummyDevice run %r (ignored)" % task)
    self.run_called_count += 1

  def set_net_params(self, network):
    print("DummyDevice set_net_params (ignored)")


class DummyDeviceBatchRun(TaskThread.DeviceBatchRun):

  def __init__(self, task="train"):
    dev = DummyDevice()
    parent = getDeviceBatchRunParent(dev=dev, task=task)
    super(DummyDeviceBatchRun, self).__init__(parent=parent, devices=[dev])

  def start(self):
    # Do nothing. Don't start the thread.
    print("DummyDeviceBatchRun: dummy start thread (ignored)")

  def allocate(self):
    self.devices_batches_idx = self.parent.batches.get_current_batch_idx()
    assert len(self.alloc_devices) == 1
    self.devices_batches = [None] * len(self.alloc_devices)
    self.num_frames = NumbersDict(13)
    batch_dim = 1
    self.alloc_devices[0].alloc_data(shapes={
      "data": (self.num_frames["data"], batch_dim, config.typed_value("num_inputs")),
      "classes": (self.num_frames["classes"], batch_dim)})
    self.parent.num_frames += self.num_frames
    self.allocated = True

  def set_dummy_dev_output(self, output, outputs_format):
    assert len(self.alloc_devices) == 1
    dev = self.alloc_devices[0]
    assert isinstance(dev, Device)
    assert dev.blocking
    dev.output = output
    dev.outputs_format = outputs_format


def test_DeviceBatchRun_outputs_format():
  # TODO: This is broken...
  return

  dev_run = DummyDeviceBatchRun(task="train")
  assert len(dev_run.alloc_devices) == 1

  # Simulate epoch start.
  trainer = dev_run.parent
  dev_run.alloc_devices[0].start_epoch_stats()
  trainer.initialize()

  # Simulate one batch.
  dev_run.allocate()
  dev_run.device_run()
  dev_run.set_dummy_dev_output(outputs_format=["cost:foo"], output=[1.42])
  dev_run.finish()

  assert_is_instance(dev_run.result, dict)
  assert_in("results", dev_run.result)
  res_outputss = dev_run.result["results"]
  assert_is_instance(res_outputss, list)
  assert_equal(len(res_outputss), len(dev_run.alloc_devices))
  res_outputs = res_outputss[0]
  assert_is_instance(res_outputs, list)
  res_outputs_format = dev_run.result["result_format"]
  assert_is_instance(res_outputs_format, list)
  res = Device.make_result_dict(res_outputs, res_outputs_format)
  assert_is_instance(res, dict)
  pprint(res)

  # Simulate epoch end.
  print("train epoch score:", trainer.score, "elapsed:", hms(trainer.elapsed))
  trainer.finalize()
  dev_run.alloc_devices[0].finish_epoch_stats()

  # Now simulate the eval.
  dev_run = DummyDeviceBatchRun(task="eval")
  assert len(dev_run.alloc_devices) == 1

  # Simulate epoch start.
  tester = dev_run.parent
  dev_run.alloc_devices[0].start_epoch_stats()
  tester.initialize()

  # Simulate one batch.
  dev_run.allocate()
  dev_run.device_run()
  dev_run.set_dummy_dev_output(outputs_format=["cost:foo", "error:foo"], output=[1.42, 2.34])
  dev_run.finish()

  # Simulate epoch end.
  print("eval epoch elapsed:", hms(tester.elapsed))
  tester.finalize()
  dev_run.alloc_devices[0].finish_epoch_stats()

  print("eval results:", tester.score, tester.error)

  assert_is_instance(dev_run.result, dict)
  assert_in("results", dev_run.result)
  res_outputss = dev_run.result["results"]
  assert_is_instance(res_outputss, list)
  assert_equal(len(res_outputss), len(dev_run.alloc_devices))
  res_outputs = res_outputss[0]
  assert_is_instance(res_outputs, list)
  res_outputs_format = dev_run.result["result_format"]
  assert_is_instance(res_outputs_format, list)
  res = Device.make_result_dict(res_outputs, res_outputs_format)
  assert_is_instance(res, dict)
  pprint(res)

  assert_greater(tester.score, 0)
  assert_greater(tester.error, 0)




import sys
import os

import _setup_test_env  # noqa
import returnn.theano.native_op as theano_native_op
import numpy as np
from numpy.testing.utils import assert_almost_equal
import theano.tensor as T
import returnn.theano.util as theano_util
f32 = "float32"


theano_util.monkey_patches()


def test_sparse_to_dense():
  n_time = 3
  n_batch = 2
  n_dim = 5
  s0 = np.array([[0,0], [0,1], [1,1], [1,2], [1,2], [2,2], [2,2]], dtype=f32)
  s1 = np.array([[1,2], [2,3], [1,1], [2,0], [4,1], [3,3], [4,4]], dtype=f32)
  w =  np.array([[1,2], [2,1], [1,2], [3,4], [5,6], [7,8], [9,9]], dtype=f32)
  m =  np.array([[1,1], [1,1], [1,1], [1,1], [1,1], [1,1], [1,0]], dtype=f32)
  W = np.array(
    [[[0, 1, 2, 0, 0], [0, 0, 2, 0, 0]],
     [[0, 1, 3, 0, 5], [0, 2, 0, 1, 0]],
     [[0, 0, 0, 7, 9], [4, 6, 0, 8, 0]]], dtype=f32)
  assert W.shape == (n_time, n_batch, n_dim)
  W2 = theano_native_op.sparse_to_dense(s0, s1, w, m, n_time, n_dim).eval()
  assert_almost_equal(W, W2)


def test_sparse_to_dense_2():
  n_time = 3
  n_batch = 2
  n_dim = 5
  s0 = np.array([[0, 0], [0, 1],  [1, 1],  [1, 2],  [1, 2],  [2, 2],  [2, 2]], dtype=f32)
  s1 = np.array([[1, 2], [2, 3],  [1, 1],  [2, 0],  [4, 1],  [3, 3],  [4, 4]], dtype=f32)
  w =  np.array([[.3,1], [.7,.4], [.1,.6], [.3,.2], [.6,.3], [.4,.5], [.6,9]], dtype=f32)
  m =  np.array([[1, 1], [1, 1],  [1, 1],  [1, 1],  [1, 1],  [1, 1],  [1, 0]], dtype=f32)
  W = np.array(
    [[[0,.3,.7, 0, 0], [0, 0, 1, 0, 0]],
     [[0,.1,.3, 0,.6], [0,.6, 0,.4, 0]],
     [[0, 0, 0,.4,.6], [.2,.3,0,.5, 0]]], dtype=f32)
  assert W.shape == (n_time, n_batch, n_dim)
  W2 = theano_native_op.sparse_to_dense(s0, s1, w, m, n_time, n_dim).eval()
  assert_almost_equal(W, W2)
  assert_almost_equal(
    np.sum(W, axis=2),
    np.ones((n_time, n_batch), dtype=f32))


def test_crossentropy_softmax_and_gradient_z_sparse():
  n_time = 3
  n_batch = 2
  n_dim = 5
  s0 = np.array([[0, 0], [0, 1],  [1, 1],  [1, 2],  [1, 2],  [2, 2],  [2, 2]], dtype=f32)
  s1 = np.array([[1, 2], [2, 3],  [1, 1],  [2, 0],  [4, 1],  [3, 3],  [4, 4]], dtype=f32)
  w =  np.array([[.3,1], [.7,.4], [.1,.6], [.3,.2], [.6,.3], [.4,.5], [.6,9]], dtype=f32)
  m =  np.array([[1, 1], [1, 1],  [1, 1],  [1, 1],  [1, 1],  [1, 1],  [1, 0]], dtype=f32)
  print("y_target:\n%r" % theano_native_op.sparse_to_dense(s0, s1, w, m, n_time, n_dim).eval())
  np.random.seed(123)
  z = np.random.randn(n_time, n_batch, n_dim).astype(f32)
  print("z:\n%r" % z)
  print("y (softmax(z)):\n%r" % theano_util.softmax(z).eval())
  z_mask = np.array([[1,1], [1,1], [1,1]], dtype=f32)
  args = (z, z_mask, s0, s1, w, m)
  ce1, gradz1 = theano_native_op.crossentropy_softmax_and_gradient_z_sparse(*args)
  ce2, gradz2 = theano_native_op.crossentropy_softmax_and_gradient_z_sparse__slow(*args)
  ce1 = ce1.eval()
  ce2 = ce2.eval()
  gradz1 = gradz1.eval()
  gradz2 = gradz2.eval()
  print("ce1:\n%r" % ce1)
  print("ce2:\n%r" % ce2)
  print("gradz1:\n%r" % gradz1)
  print("gradz2:\n%r" % gradz2)
  assert_almost_equal(ce1, ce2, decimal=5)
  assert_almost_equal(gradz1, gradz2, decimal=5)


def test_crossentropy_softmax_and_gradient_z_sparse_viterbi():
  n_time = 3
  n_batch = 2
  n_dim = 5
  alignment = np.array([[0, 1], [1, 2], [2, 3]], dtype="int32")
  mask = np.array([[1, 1], [1, 1], [1, 1]], dtype=f32)
  y_t, y_i, y_w, y_mask = theano_native_op.onehot_to_sparse(alignment, mask)
  np.random.seed(123)
  z = np.random.randn(n_time, n_batch, n_dim).astype(f32)
  z_mask = np.array([[1,1], [1,1], [1,1]], dtype=f32)
  nll1, _pcx1 = T.nnet.crossentropy_softmax_1hot(x=T.as_tensor_variable(z).reshape((n_time * n_batch, n_dim)), y_idx=T.as_tensor_variable(alignment).reshape((n_time * n_batch,)))
  nll2, _gradz2 = theano_native_op.crossentropy_softmax_and_gradient_z_sparse(z, z_mask, y_t, y_i, y_w, y_mask)
  nll1 = nll1.eval()
  nll2 = nll2.eval()
  print("nll1:\n%r" % nll1)
  print("nll2:\n%r" % nll2)


def test_max_and_argmax_sparse():
  n_time = 3
  n_batch = 2
  n_dim = 5
  s0 = np.array([[0,0], [0,1], [1,1], [1,2], [1,2], [2,2], [2,2]], dtype=f32)
  s1 = np.array([[1,2], [2,3], [1,1], [2,0], [4,1], [3,3], [4,4]], dtype=f32)
  w =  np.array([[1,2], [2,1], [1,2], [3,4], [5,6], [7,8], [9,13]], dtype=f32)
  m =  np.array([[1,1], [1,1], [1,1], [1,1], [1,1], [1,1], [1,0]], dtype=f32)
  print("W:\n%r" % theano_native_op.sparse_to_dense(s0, s1, w, m, n_time, n_dim).eval())
  init_out_max = T.zeros((n_time, n_batch), dtype=f32)
  init_out_arg = T.zeros((n_time, n_batch), dtype=f32)
  max1, arg1 = theano_native_op.max_and_argmax_sparse(s0, s1, w, m, init_out_max, init_out_arg)
  W = theano_native_op.sparse_to_dense(s0, s1, w, m, n_time, n_dim)
  assert W.ndim == 3
  max2, arg2 = T.max_and_argmax(W, axis=2)
  arg0 = np.array([[2, 2], [4, 1], [4, 3]])
  max0 = np.array([[2, 2], [5, 2], [9, 8]])
  arg1 = arg1.eval()
  arg2 = arg2.eval()
  max1 = max1.eval()
  max2 = max2.eval()
  print("arg0:\n%r" % arg0)
  print("arg1:\n%r" % arg1)
  print("arg2:\n%r" % arg2)
  print("max0:\n%r" % max0)
  print("max1:\n%r" % max1)
  print("max2:\n%r" % max2)
  assert_almost_equal(arg0, arg1)
  assert_almost_equal(arg0, arg2)
  assert_almost_equal(max0, max1)
  assert_almost_equal(max0, max2)



import numpy
import theano
import theano.gradient
import theano.tensor as T
import theano.printing
import theano.gof
from theano.sandbox.cuda.basic_ops import (as_cuda_ndarray_variable,
                                           gpu_contiguous)
from theano.gof.opt import OpSub
from theano.compile import optdb
import os

class LSTMOpCellGrad(theano.sandbox.cuda.GpuOp):
  def __init__(self, inplace):
    self.inplace = inplace
    if inplace:
      #all outputs operate inplace on inputs 4 and 6 (which are DZ and H)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys inputs 4 and 6
      #anyway theano knows that inputs 4 and 6 will be destroyed, so it should be OK
      #TODO
      #self.destroy_map = {0: [4], 1: [6]}
      #self.destroy_map = {0: [4]}
      self.destroy_map = {}

  def __eq__(self, other):
    return type(self) == type(other) and self.inplace == other.inplace

  def __str__(self):
    if self.inplace:
      return '%s{inplace}' % self.__class__.__name__
    else:
      return '%s{no_inplace}' % self.__class__.__name__

  def __hash__(self):
    return hash(type(self)) ^ hash(self.inplace)

  def make_node(self, V_h, c, idx, Dd, DY, Y, H):
    V_h = gpu_contiguous(as_cuda_ndarray_variable(V_h))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    DY = gpu_contiguous(as_cuda_ndarray_variable(DY))
    idx = gpu_contiguous(as_cuda_ndarray_variable(idx))
    Dd = gpu_contiguous(as_cuda_ndarray_variable(Dd))
    assert V_h.dtype == "float32"
    assert DY.dtype == 'float32'
    assert Y.dtype == 'float32'
    assert H.dtype == 'float32'
    assert c.dtype == 'float32'
    assert V_h.ndim == 2
    assert DY.ndim == 2
    assert Y.ndim == 2
    assert H.ndim == 2
    assert c.ndim == 2
    assert idx.ndim == 1

    return theano.Apply(self, [V_h, c, idx, Dd, DY, Y, H], [H.type(), V_h.type(), c.type()])

  #def infer_shape(self, node, input_shapes):
  #  V_hs, cs, idxs, Dds, DYs, Ys, Hs = input_shapes
  #  return [Hs, V_hs, cs]

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    V_h, c, i, Dd, DY, Y, H = input_names
    DZ, DV_h, Dc = output_names
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """

    // std::cout << "LSTMOpCellGrad called" << std::endl;
    if(!%(inplace)s)
    {
      //std::cout << "warning, inplace optimization failed, not working inplace" << std::endl;
    }

    if(%(DZ)s || %(DV_h)s || %(Dc)s)
    {
      //printf("output storage already exists\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(DZ)s);
      Py_XDECREF(%(DV_h)s);
      Py_XDECREF(%(Dc)s);
    }

    CudaNdarray * epsilon = 0;
    CudaNdarray * delta = 0;
    if(%(inplace)s)
    {
      epsilon = %(DY)s;
      delta = %(H)s;
      Py_XINCREF(delta);
    }
    else
    {
      epsilon = (CudaNdarray *) CudaNdarray_Copy(%(DY)s);
      delta = (CudaNdarray *) CudaNdarray_Copy(%(H)s);
    }

    const int * H_dim = CudaNdarray_HOST_DIMS(%(H)s);

    int y = 0;
    int x = 0;
    do_lstm_bwd(delta, epsilon, %(Y)s, %(Dd)s, %(c)s, y, x, true, %(i)s);

    %(DV_h)s = CudaNdarray_uninitialized_like(%(V_h)s);
    affine_global(%(Y)s, delta, %(DV_h)s, true, false, 1, 0.0f);

    %(DZ)s = delta;

    %(Dc)s = CudaNdarray_uninitialized_like(%(c)s);
    const int * Y_dim = CudaNdarray_HOST_DIMS(%(Y)s);
    cudaMemcpy(CudaNdarray_DEV_DATA(%(Dc)s), CudaNdarray_DEV_DATA(epsilon),
      Y_dim[0]*sizeof(float), cudaMemcpyDeviceToDevice);

    if(!%(inplace)s)
    {
      Py_XDECREF(epsilon);
    }

    """ % locals()

  #!!! change this when changing the code!
  #def c_code_cache_version(self):
  #  return 1, 2

LSTMOpCellGradNoInplaceInstance = LSTMOpCellGrad(inplace=False)
LSTMOpCellGradInplaceInstance = LSTMOpCellGrad(inplace=True)

LSTMOpCellGradInplaceOpt = OpSub(LSTMOpCellGradNoInplaceInstance, LSTMOpCellGradInplaceInstance)

#hack to avoid being called twice
if not hasattr(optdb, 'LSTMOpCellGradInplaceOpt_registered'):
  optdb.register('LSTMOpCellGradInplaceOpt', theano.gof.TopoOptimizer(LSTMOpCellGradInplaceOpt),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.LSTMOpCellGradInplaceOpt_registered = True


#------------------------

class LSTMOpCell(theano.sandbox.cuda.GpuOp):
  def __init__(self, inplace):
    self.inplace = inplace
    if inplace:
      #all outputs operate inplace on input 0 (which is Z)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys input 0
      #anyway theano knows that input 0 will be destroyed, so it should be OK
      #TODO
      #self.destroy_map = {0: [0]}
      self.destroy_map = {}

  def __eq__(self, other):
    return type(self) == type(other) and self.inplace == other.inplace

  def __str__(self):
    if self.inplace:
      return '%s{inplace}' % self.__class__.__name__
    else:
      return '%s{no_inplace}' % self.__class__.__name__

  def __hash__(self):
    return hash(type(self)) ^ hash(self.inplace)

  def make_node(self, Z, V_h, c, i):
    Z = gpu_contiguous(as_cuda_ndarray_variable(Z))
    V_h = gpu_contiguous(as_cuda_ndarray_variable(V_h))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    i = gpu_contiguous(as_cuda_ndarray_variable(i))
    assert Z.dtype == "float32"
    assert V_h.dtype == "float32"
    assert c.dtype == 'float32'
    assert c.ndim == 2
    assert Z.ndim == 2
    assert i.ndim == 1
    assert V_h.ndim == 2

    #results: output Y, (gates and cell state) H
    return theano.Apply(self, [Z, V_h, c, i], [Z.type(), Z.type(), c.type()])

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    X, V_h, c, i = input_names
    Z, H, d = output_names
    fail = sub['fail']
    return """
    if(%(Z)s || %(H)s || %(d)s)
    {
      //printf("Z or H already exist\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(Z)s);
      Py_XDECREF(%(H)s);
      Py_XDECREF(%(d)s);
    }

    const int * X_dim = CudaNdarray_HOST_DIMS(%(X)s);
    //we can't use the modulo operator easily as it should not be replaced
    const int dims_Z[] = {X_dim[0], X_dim[1] / 4};
    const int dims_H[] = {X_dim[0], X_dim[1]};
    const int dims_d[] = {X_dim[0], X_dim[1] / 4};
    int size_d = X_dim[0] * X_dim[1] / 4;

    %(Z)s = (CudaNdarray*) CudaNdarray_NewDims(2,dims_Z);
    %(d)s = (CudaNdarray*) CudaNdarray_NewDims(2, dims_d);
    %(H)s = (CudaNdarray*) CudaNdarray_NewDims(2,dims_H); //CudaNdarray_uninitialized_like(%(X)s);
    cudaMemcpy(CudaNdarray_DEV_DATA(%(H)s), CudaNdarray_DEV_DATA(%(X)s),
      dims_H[0]*dims_H[1]*sizeof(float), cudaMemcpyDeviceToDevice);

    int y = 0;
    int x = 0;
    if(x > 0)
    {
      //H += Z[x-1]*V_h
      affine_y_x(y, x-1, %(Z)s, y, x, %(V_h)s, y, x, %(H)s);
    }
    float * d_ptr =  CudaNdarray_DEV_DATA(%(d)s);
    do_lstm(%(H)s, %(Z)s, %(c)s, d_ptr, y, x, %(i)s);
    """ % locals()

  def grad(self, inputs, output_grads):
    Z, V_h, c, i = inputs
    DY, DH, Dd = output_grads

    Z_raw = Z.owner.inputs[0].owner.inputs[0]
    #TODO!!!
    V_h_raw = V_h.owner.inputs[0]
    c_raw = c.owner.inputs[0].owner.inputs[0]
    i_raw = i.owner.inputs[0].owner.inputs[0]
    #we have to make sure that this in only computed once!
    #for this we have to extract the raw variables before conversion to continuous gpu array
    #so that theano can merge the nodes
    Z, H, d = LSTMOpCellInstance(Z_raw, V_h_raw, c_raw, i_raw)
    if isinstance(DY.type, theano.gradient.DisconnectedType):
      DY = T.zeros_like(Z)
    if isinstance(Dd.type, theano.gradient.DisconnectedType):
      Dd = T.zeros_like(c)
    DZ, DV_h, Dc = LSTMOpCellGradNoInplaceInstance(V_h, c, i, Dd, DY, Z, H)
    Di = theano.gradient.grad_undefined(self, 3, inputs[3], 'cannot diff w.r.t. index')

    return [DZ, DV_h, Dc, Di]

  def infer_shape(self, node, input_shapes):
    Xs, V_hs, cs, idxs = input_shapes
    Z_shape = (Xs[0], Xs[1] / 4)
    H_shape = (Xs[0], Xs[1])
    D_shape = (Xs[0], Xs[1] / 4)
    return [Z_shape, H_shape, D_shape]

  #!!! change this when changing the code!
  #def c_code_cache_version(self):
  #  return 1, 2

LSTMOpCellInstance = LSTMOpCell(inplace=False)
LSTMOpCellInplaceInstance = LSTMOpCell(inplace=True)

LSTMOpCellInplaceOpt = OpSub(LSTMOpCellInstance, LSTMOpCellInplaceInstance)

#hack to avoid begin aclled twice
if not hasattr(optdb, 'LSTMOpCellInplaceOpt_registered'):
  optdb.register('LSTMOpCellInplaceOpt', theano.gof.TopoOptimizer(LSTMOpCellInplaceOpt),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.LSTMOpCellInplaceOpt_registered = True



#!/usr/bin/env python3

"""
Dump translation dataset to HDF.
"""

from __future__ import print_function

import sys
import argparse
import gzip
import pickle
import typing

import numpy
import h5py
from operator import itemgetter
from itertools import islice, zip_longest

UNKNOWN_LABEL = "<UNK>"
POSTFIX = "</S>"


class HDFTranslationDatasetCreator(object):
  """
  Creates the same HDF file as hdf_dump.py called on a TranslationDataset, but does so much faster
  and using much less memory.

  Input parameters are basically the files in a TranslationDataset folder. Additionally, the number of lines has to
  be given.
  """

  def __init__(self, hdf_file, source_file, target_file, source_vocabularies, target_vocabularies,
               source_factors, target_factors, number_of_lines, factor_separator="|",
               compression_method=None, line_buffer_size=100000, data_buffer_size=5000000):
    """
    :param str hdf_file: filename for the hdf file being created
    :param str source_file: filename of the source text file
    :param str target_file: filename of the target text file
    :param list[str] source_vocabularies: Filenames of the source vocabularies (in pickle format).
      Usually there is just one. In case of source factors, provide one per factor.
    :param list[str] target_vocabularies: Filenames of the target vocabularies (in pickle format).
      Usually there is just one. In case of target factors, provide one per factor.
    :param list[str] source_factors: Data keys for the source factors. First data key is always 'data'
      and must not be contained in the list.
    :param list[str] target_factors: Data keys for the target factors. First data key is always 'classes'
      and must not be contained in the list.
    :param str factor_separator: "String used to separate factors of the words. E.g. if "|", words are expected to be
      of format "<lemma>|<factor>|...".
    :param int number_of_lines: line count in source_file and target_file
    :param str compression_method: Optional compression method as supported by h5py.File.create_dataset().
        Applied to the main data only ('inputs' and 'target/data/classes').
    :param int line_buffer_size: number of corpus lines to read and process at once
    :param int data_buffer_size: space to reserve inside the hdf file at once, in numbers of integers
    """

    self.hdf_file_name = hdf_file
    self.hdf_file = None
    self.compression_method = compression_method

    self.source_file_handle = self._open_file(source_file)
    self.target_file_handle = self._open_file(target_file)

    source_vocabularies = [self._read_vocabulary(v) for v in source_vocabularies]
    target_vocabularies = [self._read_vocabulary(v) for v in target_vocabularies]

    self.source_data_keys = ["data"] + source_factors
    self.target_data_keys = ["classes"] + target_factors

    self._vocabularies = {"source": source_vocabularies, "target": target_vocabularies}
    self._vocabulary_sizes = {
      "source": [len(v) for v in source_vocabularies],
      "target": [len(v) for v in target_vocabularies]}
    self._unknown_ids = {
      "source": [v.get(UNKNOWN_LABEL) for v in source_vocabularies],
      "target": [v.get(UNKNOWN_LABEL) for v in target_vocabularies]}

    self.number_of_lines = number_of_lines
    self.line_buffer_size = line_buffer_size
    self.data_buffer_size = data_buffer_size

    self._number_of_processed_lines = 0
    self._write_offsets = {data_key: 0 for data_key in self.source_data_keys + self.target_data_keys}

    self.factor_separator = factor_separator

  def create(self):
    """
    Main function writing the HDF file.
    """
    self._init_hdf_file()

    print("Setting attributes...", file=sys.stderr)
    sys.stderr.flush()
    self._write_attributes()
    print("Done.", file=sys.stderr)

    print("Writing labels (vocabulary)...", file=sys.stderr)
    sys.stderr.flush()
    self._write_labels()
    print("Done.", file=sys.stderr)

    print("Writing source, target, sequence lengths and tags for all lines...", file=sys.stderr)
    sys.stderr.flush()
    end_of_file = False
    while not end_of_file:
      end_of_file = self._write_data()
      if not end_of_file:
        print("> Processed {} lines.".format(self._number_of_processed_lines), file=sys.stderr)
        sys.stderr.flush()
    print("Done.", file=sys.stderr)
    sys.stderr.flush()

    self.hdf_file.close()

  def _init_hdf_file(self):
    """
    Sets up the HDF file and initializes the datasets that will be filled.
    """
    self.hdf_file = h5py.File(self.hdf_file_name, "w")

    self.hdf_file.create_group("targets/data")
    self.hdf_file.create_group("targets/size")
    self.hdf_file.create_group("targets/labels")

    num_data_keys = len(self.source_data_keys) + len(self.target_data_keys)
    self.hdf_file.create_dataset("seqLengths", (self.number_of_lines, num_data_keys), dtype="int32")
    max_tag_length = len("line-") + len(str(self.number_of_lines))
    self._tag_dtype = "S{}".format(max_tag_length)
    self.hdf_file.create_dataset("seqTags", (self.number_of_lines,), dtype=self._tag_dtype)

    self.hdf_file.create_dataset("inputs", (self.data_buffer_size,), maxshape=(None,),
                                 dtype="int32", compression=self.compression_method)

    # HDF format expects one input only, so store source factors as target too.
    for data_key in self.source_data_keys[1:] + self.target_data_keys:
      self.hdf_file["targets/data"].create_dataset(data_key, (self.data_buffer_size,), maxshape=(None,),
                                                   dtype="int32", compression=self.compression_method)

  def _write_attributes(self):
    """
    Writes several attributes to the HDF file.
    """
    for index, data_key in enumerate(self.source_data_keys[1:], start=1):
      self.hdf_file["targets/size"].attrs[data_key] = (self._vocabulary_sizes["source"][index], 1)
    for index, data_key in enumerate(self.target_data_keys):
      self.hdf_file["targets/size"].attrs[data_key] = (self._vocabulary_sizes["target"][index], 1)

    # Those should be deprecated, but include nevertheless to exactly reproduce hdf_dump.
    self.hdf_file.attrs["inputPattSize"] = self._vocabulary_sizes["source"][0]
    self.hdf_file.attrs["numLabels"] = self._vocabulary_sizes["target"][0]

  def _write_labels(self):
    """
    Writes the labels (i.e. target vocabulary) to the HDF file.
    """
    for side in ["source", "target"]:
      # We have to write it for the source factors too, because they are treated like targets.

      data_keys = self.source_data_keys if side == "source" else self.target_data_keys
      for index, data_key in enumerate(data_keys):
        if data_key == "data":
          continue

        sorted_vocabulary_tuples = sorted(self._vocabularies[side][index].items(), key=itemgetter(1))
        labels = [word.encode("utf8") for (word, _) in sorted_vocabulary_tuples]

        assert len(labels) == self._vocabulary_sizes[side][index], "Word ids were not unique."

        max_label_length = max([len(label) for label in labels])
        dtype = "S{}".format(max_label_length + 1)

        labels = [numpy.array(label, dtype=dtype, ndmin=1) for label in labels]
        labels = numpy.concatenate(labels)

        self.hdf_file["targets/labels"].create_dataset(
          data_key, (self._vocabulary_sizes[side][index],),
          data=labels, dtype=dtype)

  def _write_data(self):
    """
    Loads a chunk of lines from the corpus and writes all corresponding data to the HDF file.

    :return: whether the end of the corpus is reached
    :rtype: bool
    """
    data_chunks = self._get_chunk()

    if not data_chunks["data"]:
      self._finalize_data()
      return True

    source_lengths = [len(line) for line in data_chunks["data"]]
    target_lengths = [len(line) for line in data_chunks["classes"]]

    chunk_size = len(data_chunks["data"])

    self._write_lengths(source_lengths, target_lengths)
    self._write_tags(chunk_size)

    for data_key, data_chunk in data_chunks.items():
      self._write_data_indices(data_chunk, data_key)

    self._number_of_processed_lines += chunk_size

    return False

  def _get_chunk(self):
    """
    Reads in the next chunk of lines from the corpus files.

    :return: a dict 'data_key' -> word indices for 'data_key' (int32, shape [Lines]) for all source and target data
    :rtype: dict[str,list[numpy.ndarray]]
    """
    source_lines = islice(self.source_file_handle, self.line_buffer_size)
    target_lines = islice(self.target_file_handle, self.line_buffer_size)

    data_chunks = {data_key: [] for data_key in self.source_data_keys + self.target_data_keys}
    for source_line, target_line in zip_longest(source_lines, target_lines):
      assert source_line is not None and target_line is not None, "Number of source and target lines differ."

      source_indices = self._line_to_indices(source_line, "source")
      for data_key, indices in zip(self.source_data_keys, source_indices):
        data_chunks[data_key].append(indices)

      target_indices = self._line_to_indices(target_line, "target")
      for data_key, indices in zip(self.target_data_keys, target_indices):
        data_chunks[data_key].append(indices)

    return data_chunks

  def _write_lengths(self, source_lengths, target_lengths):
    """
    Writes the sequence lengths to the HDF file.

    :param list[int] source_lengths: lengths of all source lines in current chunk
    :param list[int] target_lengths: lengths of all target lines in current chunk
    """
    # We treat source factors as targets internally because HDF format does not support multiple inputs.

    # For each sequence, seqLengths is expected to contain the length of the input and the length of each of the
    # targets, in alphabetical order of the target data keys. As all source and all target factors share the
    # same lengths we just choose between the source and target lengths.
    target_sequence_lengths = {}
    for data_key in self.source_data_keys[1:]:
      target_sequence_lengths[data_key] = source_lengths
    for data_key in self.target_data_keys:
      target_sequence_lengths[data_key] = target_lengths

    # Now sort by key.
    key_lengths_tuples_sorted = sorted(target_sequence_lengths.items(), key=lambda x: x[0])  # type: typing.List[typing.Tuple[str,typing.List[int]]]  # nopep8
    target_lengths_sorted = [key_length_tuple[1] for key_length_tuple in key_lengths_tuples_sorted]

    # Finally, add one time the source lengths for the input ("data") and convert to numpy.
    lengths = numpy.array([source_lengths] + target_lengths_sorted, dtype="int32").transpose()

    offset = self._number_of_processed_lines
    assert len(lengths) + offset <= self.number_of_lines, "More lines in the corpus files than specified."

    self.hdf_file["seqLengths"][offset:offset + len(lengths), :] = lengths

  def _write_tags(self, chunk_size):
    """
    Writes the sequence tags to the HDF file.

    :param int chunk_size: number of lines in the current chunk
    """
    offset = self._number_of_processed_lines

    tags = [numpy.array("line-" + str(offset + i), dtype=self._tag_dtype, ndmin=1) for i in range(chunk_size)]
    tags = numpy.concatenate(tags)

    self.hdf_file["seqTags"][offset: offset + chunk_size] = tags

  def _write_data_indices(self, chunk, data_key):
    """
    Writes the main data (word indices for the source or target corpus) to the HDF file.

    :param list[numpy.ndarray] chunk: word indices for all lines in the current chunk
    :param str data_key: "data", "classes" or a name of a factor
    """
    indices = numpy.concatenate(chunk)

    if data_key == "data":
      dataset = self.hdf_file["inputs/"]
    else:
      dataset = self.hdf_file["targets/data/{}".format(data_key)]

    offset = self._write_offsets[data_key]
    length = len(indices)
    if offset + length > len(dataset):
      buffer_size = max(self.data_buffer_size, length)
      dataset.resize((offset + buffer_size,))

    dataset[offset:offset + length] = indices

    self._write_offsets[data_key] += length

  def _finalize_data(self):
    """
    Called after all data is written. Checks number of lines and resizes datasets down to actual data size.
    """
    # Make sure the number of lines given by the user was correct.
    # Otherwise lengths and labels would have trailing zeros.
    assert self.number_of_lines == self._number_of_processed_lines, (
      "Fewer lines ({}) in the corpus files "
      "than specified ({}).".format(self._number_of_processed_lines, self.number_of_lines))

    # Trim datasets to actually occupied length, i.e. remove unused reserved space.
    self.hdf_file["inputs"].resize((self._write_offsets["data"],))
    for data_key in self.source_data_keys[1:] + self.target_data_keys:
      self.hdf_file["targets/data/{}".format(data_key)].resize((self._write_offsets[data_key],))

  def _line_to_indices(self, line, side):
    """
    Converts a line of text to arrays of word indices.

    :param str line: input line
    :param str side: "source" or "target"
    :return: word indices (int32, shape [num_words]) for all source or target factors
    :rtype: list[numpy.ndarray]
    """
    data_keys = self.source_data_keys if side == "source" else self.target_data_keys

    words = line.strip().split()

    if len(data_keys) == 1:
      word_list_per_factor = [words + [POSTFIX]]
    else:
      if words:
          words_split_into_factors = [word.split(self.factor_separator) for word in words]
          assert all(len(factors) == len(data_keys) for factors in words_split_into_factors), (
            "All words must have all factors. Expected: " + self.factor_separator.join(data_keys))
          word_list_per_factor = zip(*words_split_into_factors)
      else:
          word_list_per_factor = [[]] * len(data_keys)
      word_list_per_factor = [list(words) + [POSTFIX] for words in word_list_per_factor]

    indices_list = []
    for index, (data_key, words) in enumerate(zip(data_keys, word_list_per_factor)):
      indices = [self._vocabularies[side][index].get(word, self._unknown_ids[side][index]) for word in words]

      indices_numpy = numpy.array(indices, dtype=numpy.int32)
      indices_list.append(indices_numpy)

    return indices_list

  @staticmethod
  def _open_file(file_name):
    """
    :param str file_name: filename of a plain text file, possibly zipped
    :return: file handle
    :rtype: io.TextIOWrapper|gzip.GzipFile
    """
    if file_name.endswith(".gz"):
      return gzip.open(file_name, "rt")
    else:
      return open(file_name, "r")

  @staticmethod
  def _read_vocabulary(file_name):
    """
    :param str file_name: filename of the vocabulary (in pickle format)
    :return: mapping from words to indices
    :rtype: dict[str,int]
    """
    file_handle = open(file_name, "rb")
    vocabulary = pickle.load(file_handle)
    return vocabulary


def parse_args():
  """
  :rtype: argparse.Namespace
  """
  parser = argparse.ArgumentParser()
  parser.add_argument("-s", "--source_corpus", required=True, help="Source corpus file, possibly zipped.")
  parser.add_argument("-t", "--target_corpus", required=True, help="Target corpus file, possibly zipped.")
  parser.add_argument(
    "-v", "--source_vocabulary", required=True, help="Source vocabulary in pickle format."
    "In case of source factors provide a comma separated list containing vocabularies for each factor.")
  parser.add_argument(
    "-w", "--target_vocabulary", required=True, help="Target vocabulary in pickle format."
    "In case of target factors provide a comma separated list containing vocabularies for each factor.")
  parser.add_argument("-o", "--hdf_file", required=True, help="Output HDF file name.")
  parser.add_argument("--source_factors", help="Comma separated list of data keys for the source factors.")
  parser.add_argument("--target_factors", help="Comma separated list of data keys for the target factors.")
  parser.add_argument(
    "--factor_separator", default="|",
    help="String used to separate factors of the words, E.g. if '|', words are expected to be "
    "of format '<lemma>|<factor>|...'")
  parser.add_argument(
    "-n", "--number_of_lines", required=True, type=int,
    help="The number of total lines in the corpus files.")
  parser.add_argument("-c", "--compression", help="Type of compression (e.g. 'gzip', 'lzf'). Turned off if not given.")
  parser.add_argument("-l", "--line_buffer_size", type=int, help="How many lines to read at once.", default=100000)
  parser.add_argument(
    "-d", "--data_buffer_size", type=int, help="How much space to reserve in the HDF dataset "
    "at once (in number of integers).", default=5000000)

  return parser.parse_args()


def main():
  """
  Main entry.
  """
  args = parse_args()

  # In case of source or target factors we need a vocabularies for each.
  source_vocabularies = args.source_vocabulary.split(",")
  target_vocabularies = args.target_vocabulary.split(",")

  source_factors = args.source_factors.split(",") if args.source_factors else []
  target_factors = args.target_factors.split(",") if args.target_factors else []

  assert len(source_factors) + 1 == len(source_vocabularies), (
    "Number of source factors must be one less "
    "than number of source vocabularies (first factor is always called 'data')")
  assert len(target_factors) + 1 == len(target_vocabularies), (
    "Number of target factors must be one less "
    "than number of target vocabularies (first factor is always called 'classes')")

  HDFTranslationDatasetCreator(
    args.hdf_file, args.source_corpus, args.target_corpus,
    source_vocabularies, target_vocabularies, source_factors, target_factors, args.number_of_lines,
    args.factor_separator,
    args.compression, args.line_buffer_size, args.data_buffer_size).create()


if __name__ == "__main__":
  sys.exit(main())



#!/usr/bin/env python

"""
Iterate through a dataset, just like RETURNN would do in training.
"""

from __future__ import print_function

import sys
import typing

import _setup_returnn_env  # noqa
from returnn import __main__ as rnn
from returnn.log import log
from returnn.engine.base import EngineBase

dev_num_batches = 1


def iterate_dataset(dataset, recurrent_net, batch_size, max_seqs):
  """
  :type dataset: Dataset.Dataset
  :type recurrent_net: bool
  :type batch_size: int
  :type max_seqs: int
  """
  batch_gen = dataset.generate_batches(recurrent_net=recurrent_net, batch_size=batch_size, max_seqs=max_seqs)
  while batch_gen.has_more():
    batches = batch_gen.peek_next_n(dev_num_batches)
    for batch in batches:
      dataset.load_seqs(batch.start_seq, batch.end_seq)
    batch_gen.advance(len(batches))


def iterate_epochs():
  """
  Iterate through epochs.
  """
  start_epoch, start_batch = EngineBase.get_train_start_epoch_batch(config)
  final_epoch = EngineBase.config_get_final_epoch(config)

  print("Starting with epoch %i, batch %i." % (start_epoch, start_batch), file=log.v3)
  print("Final epoch is: %i" % final_epoch, file=log.v3)

  recurrent_net = "lstm" in config.value("hidden_type", "")  # good enough...
  batch_size = config.int('batch_size', 1)
  max_seqs = config.int('max_seqs', -1)

  for epoch in range(start_epoch, final_epoch + 1):
    print("Epoch %i." % epoch, file=log.v3)
    rnn.train_data.init_seq_order(epoch)
    iterate_dataset(rnn.train_data, recurrent_net=recurrent_net, batch_size=batch_size, max_seqs=max_seqs)

  print("Finished all epochs.", file=log.v3)


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, command_line_options):
  """
  :param str config_filename:
  :param list[str] command_line_options:
  """
  rnn.init_better_exchook()
  rnn.init_thread_join_hack()
  rnn.init_config(config_filename, command_line_options)
  global config
  config = rnn.config
  rnn.init_log()
  print("RETURNN demo-dataset starting up", file=log.v3)
  rnn.init_faulthandler()
  rnn.init_config_json_network()
  rnn.init_data()
  rnn.print_task_properties()


def main(argv):
  """
  Main entry.
  """
  assert len(argv) >= 2, "usage: %s <config>" % argv[0]
  init(config_filename=argv[1], command_line_options=argv[2:])
  iterate_epochs()
  rnn.finalize()


if __name__ == '__main__':
  main(sys.argv)
